/************************************************************************************\

  rtstepper.h - rt-stepper dongle support for EMC2

  (c) 2011-2015 Copyright Eckler Software

  Author: David Suffield, dsuffiel@ecklersoft.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of version 2 of the GNU General Public License as published by
  the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

  Upstream patches are welcome. Any patches submitted to the author must be 
  unencumbered (ie: no Copyright or License).

  See project revision history the "configure.ac" file.

\************************************************************************************/

#ifndef _RTSTEPPER_H
#define _RTSTEPPER_H

#include <stdint.h>
#include <pthread.h>
#include <libusb.h>
#include "list.h"
#include "emc.h"

enum RTSTEPPER_IO_TYPE
{
   RTSTEPPER_IO_TYPE_SPINDLE_ASYNC,  /* feed is NOT synchronized to spindle */
   RTSTEPPER_IO_TYPE_SPINDLE_SYNC,   /* feed is synchronized to spindle */
};

struct rtstepper_io_req
{
   int id;
   EmcPose position;            // commanded position
   unsigned char *buf;          /* step/direction buffer */
   int buf_size;                /* buffer size in bytes */
   int total;                   /* current buffer count, number of bytes used (total < buf_size) */
   enum RTSTEPPER_IO_TYPE type;
   struct emc_session *session;
   struct libusb_transfer *req; 
   struct list_head list;
};

/* Dongle board FW releases. */
enum RTSTEPPER_BRD
{
   RTSTEPPER_BRD_NA,
   RTSTEPPER_BRD_1a,    /* new */
   RTSTEPPER_BRD_1b,    /* Microsoft OS descriptors */
   RTSTEPPER_BRD_1c,    /* OUTPUT0-1, usb suspend */
   RTSTEPPER_BRD_1d,    /* INPUT0 frequency counter */
   RTSTEPPER_BRD_1e,    /* index pulse threading */
   RTSTEPPER_BRD_NEW,
};

struct rtstepper_file_descriptor
{
   libusb_device_handle *hd;
   libusb_device **list_all;
   libusb_context *ctx;
   libusb_device *dev;
   int event_done;
   int event_abort_done;
   pthread_t event_tid;    /* thread handle */
   int dongle_done;
   int dongle_abort_done;
   pthread_t dongle_tid;    /* thread handle */
   enum RTSTEPPER_BRD board_rev;
};

struct rtstepper_moving_avg
{
   int sum;
   int array[4];
   int pos;      /* pointer to array[] */
};

#define RTSTEPPER_ICOUNT_MAX 0xffff      /* max value for icount_period */
#define RTSTEPPER_STEP_CLOCK 46875

#define RTSTEPPER_STEP_STATE_ABORT_BIT 0x01     /* abort step buffer, 1=True, 0=False (R/W) */
#define RTSTEPPER_STEP_STATE_EMPTY_BIT 0x02     /* step buffer empty, 1=True, 0=False */
#define RTSTEPPER_STEP_STATE_SYNC_START_BIT 0x04     /* 1=True, 0=False */
#define RTSTEPPER_STEP_STATE_INPUT0_BIT 0x08    /* active high INPUT0, 1=True, 0=False */
#define RTSTEPPER_STEP_STATE_INPUT1_BIT 0x10    /* active high INPUT1, 1=True, 0=False */
#define RTSTEPPER_STEP_STATE_INPUT2_BIT 0x20    /* active high INPUT2, 1=True, 0=False */
#define RTSTEPPER_STEP_STATE_OUTPUT0_BIT 0x40    /* 1=True, 0=False */
#define RTSTEPPER_STEP_STATE_OUTPUT1_BIT 0x80    /* 1=True, 0=False */

#define RTSTEPPER_MECH_THREAD 1
#define RTSTEPPER_DONGLE_THREAD 0

/* IO request hysteresis set points. */
#define RTSTEPPER_REQ_MAX  100
#define RTSTEPPER_REQ_MIN  50

/* Forward declarations. */
struct emc_session;

#ifdef __cplusplus
extern "C"
{
#endif

/* Function prototypes */

   enum EMC_RESULT rtstepper_open(struct emc_session *ps);
   enum EMC_RESULT rtstepper_close(struct emc_session *ps);
   enum EMC_RESULT rtstepper_state_query(struct emc_session *ps);
   enum EMC_RESULT rtstepper_encode(struct emc_session *ps, struct rtstepper_io_req *io, double index[]);
   enum EMC_RESULT rtstepper_xfr_start(struct emc_session *ps, struct rtstepper_io_req *io, EmcPose pos);
   enum EMC_RESULT rtstepper_xfr_wait(struct emc_session *ps);
   enum EMC_RESULT rtstepper_xfr_hysteresis(struct emc_session *ps);
   enum EMC_RESULT rtstepper_dongle_empty_wait(struct emc_session *ps);
   enum EMC_RESULT rtstepper_dongle_sync_start_set(struct emc_session *ps);
   enum EMC_RESULT rtstepper_dongle_sync_start_wait(struct emc_session *ps);
   enum EMC_RESULT rtstepper_clear_stats(struct emc_session *ps);
   int rtstepper_is_connected(struct emc_session *ps);
   enum EMC_RESULT rtstepper_is_input0_triggered(struct emc_session *ps);
   enum EMC_RESULT rtstepper_is_input1_triggered(struct emc_session *ps);
   enum EMC_RESULT rtstepper_is_input2_triggered(struct emc_session *ps);
   enum EMC_RESULT rtstepper_input0_state(struct emc_session *ps);
   enum EMC_RESULT rtstepper_input1_state(struct emc_session *ps);
   enum EMC_RESULT rtstepper_input2_state(struct emc_session *ps);
   enum EMC_RESULT rtstepper_home(struct emc_session *ps);
   enum EMC_RESULT rtstepper_position_set(struct emc_session *ps, EmcPose pos);
   enum EMC_RESULT rtstepper_estop(struct emc_session *ps, int thread);
   struct rtstepper_io_req *rtstepper_io_req_alloc(struct emc_session *ps, int id, enum RTSTEPPER_IO_TYPE io_type);
   enum EMC_RESULT rtstepper_test(const char *snum);
#ifdef __cplusplus
}
#endif

#endif                          /* _RTSTEPPER_H */
